# Ruby script to draw RiverArch outlines
# v 3.3 JWM 2016-01-30
# Use smaller slice height at top
# Draw all four faces from one set of inputs 
# Designed to draw one quarter of sloped ellipses with nodes at specified vertical interval - e.g., 5ft.
# initially centred at origin, as a component

# Input parameters:
# Height = semi major axis = a
# Width = semi minor axis = b
# slope = angle of plane of ellipse from vertical = slope

# Formula for ellipse:
# Given semi major and minor axes a and b: (http://www.nabla.hr/PC-Ellipse2.htm)
# Standard equation of ellipse is
# x**2/a**2 + y**2/b**2 = 1 (or in our case, we'll use z instead of xm=, and x instead of y)
# and we want to calculate x for given z.
# Rewrite therefore as
# z**2/a**2  +  x**2 / b**2 = 1

# or, solving for x

# (z**2 * b**2)/a**2 + x**2 =1

# x**2 = b**2 - (z**2 * b**2)/a**2

# x = sqrt(b**2  - (z**2 * b**2)/a**2)

# so dividing by b**2
# x**2 = a**2 - (a**2 / b**2) * z**2

# so (before taking account of the slope)
# x = sqrt(a**2 - z**2 * (a**2 / b**2))
# x = b * sqrt(1 - z**2 / a**2)

def initialize()
  # for arches
  # @defaults = [1000.feet, 1000.feet, 0.0, 875.feet, 750.feet, 0.0,  0.0.feet, 10.feet, "Arch  ", "No" ]
  # for internal elevators
  #  @defaults = [999.feet, 995.feet, 0.0, 999.feet, 995.feet, 0.0,  6.feet, 10.feet, "Column outer face", "No" ]
  # for external elevator support and tracks
  @defaults = [893.feet, 768.feet, 0.0, 890.feet, 765.feet, 0.0,  0.feet, "Ext elevator", "Yes" ]

end

# def slices_for_floor(floor_no, std_floor_height, slice_ht)
#   non_std_floor_height = {1 => 9.83333.feet, 2 => 15.25.feet, 4 => 15.25.feet, 33 => 15.25.feet, 64 => 15.25.feet, 87  => 20.feet}
#   if !non_std_floor_height[floor_no].nil? # non-standard floor height
#   # divmod returns whole number of slices in floor, and the excess for the last slice
#      non_std_floor_height[floor_no].divmod(slice_ht)
#   else
#     [std_floor_height.divmod(slice_ht)[0], 0.0] # std no of slices, no excess in last slice
#   end # case
# end

def draw_floor(floor, p1, p2, show_floor_numbers, comp, numbers_comp)
  # puts " floor = #{floor},p1 = #{p1}, p2 = #{p2}"
  comp.entities.add_line p1, p2
  if show_floor_numbers
    text_offset = - 10.feet
    floor_no_text = numbers_comp.entities.add_text floor.to_s, p1, [text_offset,0,7.5.feet]
    floor_no_text.leader_type = 0
    floor_no_text.arrow_type = 0
  end
end

def get_inputs()
  ## Get user input parameters for arch section#######################
  defaults = @defaults
  # With four params, it shows all text boxes:
  prompts = ["Major semi-axis 1           ", "Minor semi-axis 1         ", "Slope 1 (between -1 and +1)         ",
    "Major semi-axis 2         ", "Minor semi-axis 2         ", "Slope 2 (between -1 and +1)         ",
    "y_offset",  "Name", "Add floor numbers?"]
  # for arches
  # list = ["1000'|999'|995'|950'|939'|890'|881.1667'|876'|875'","1000'|995'|890'|881.1667'|876'|875'|870'|765'|755'|751'|750'",
  #   "0.2|0.0",
  #    "1000'|999'|995'|939'|890'|890'|881.1667'|876'|875'","1000'|995'|890'|881.1667'|876'|875'|870'|765'|755'|751'|750'",
  #          "0.2|0.0","0'|3'|4'|6'|21'|389'|400'","","No|Yes"]

  # for internal elevator shafts
  #  list = ["949'|938'|937'|926'|886'6|875'6|874'6|863'6","949'|938'|937'|926'|886'6|875'6|874'6|863'6",
  #  "0.2|0.0",
  #   "949'|938'|937'|926'|886'6|875'6|874'6|863'6","949'|938'|937'|926'|886'6|875'6|874'6|863'6",
  #         "0.2|0.0","0|10'2","","No|Yes"]

  # for main structural elliptical columns
  #   list = ["|980'|976'|960'|956'|940'|936'|920'|916'|900'|896'|880'|879'|876'|874'|999'|995'|958'|954'|918'|914'|878'|874'|838'|834'|798'|794'|755'|754'|751'",  "980'|976'|960'|956'|940'|936'|920'|916'|900'|896'|880'|879'|876'|874'|999'|995'|958'|954'|918'|914'|878'|876'|874'|838'|834'|798'|794'|755'|754'|751'","0.2|0.0", "980'|976'|960'|956'|940'|936'|920'|916'|900'|896'|880'|879'|876'|999'|995'|958'|954'|918'|914'|879'|878'|876'|875'|874'|838'|834'|798'|794'|755'|754'|751'",  "980'|976'|960'|956'|940'|936'|920'|916'|900'|896'|880'|876'|999'|995'|958'|954'|918'|914'|878'|874'|838'|834'|798'|794'|755'|754'|751'","0.2|0.0","0|4'|6'|10'2","","No|Yes"]

  # For external elevator and track supports
  list = ["893'|890'|888'8|768'|765'|763'8",  "893'|890'|888'8|768'|765'|763'8","0.2|0.0", "893'|890'|888'8|768'|765'|763'8", "893'|890'|888'8|768'|765'|763'8","0.2|0.0","0|2'|21'","","No|Yes"]

  input = UI.inputbox(prompts, defaults, list, "Arch section parameters" )
  
  if defaults.nil? || defaults.empty?  # not previously defined defaults or input
    defaults = @defaults
    if !input.empty?
      # remember inputs ready to adapt for next section of arch
      @defaults = input
    end
  end

  if input
    a1 = input[0].to_l
    b1 = input[1].to_l
    slope1 = input[2].to_f
    a2 = input[3].to_l
    b2 = input[4].to_l
    slope2 = input[5].to_f
    y_offset = input[6].to_l
    component_name = input[7]
    show_floor_numbers = input[8] == "No"? false : true

    if slope1.abs >1 || slope2.abs >1
      UI.messagebox "Slope must be between -1 and +1. Please try again", MB_OK
    end
    if b1 > a1 || b2 > a2
      UI.messagebox "Major axis must be larger than or equal to minor axis", MB_OK
    end
  else
    UI.messagebox "Sorry, invalid values, please try again", MB_OK
  end
  @defaults = input
  # Find the higher apex and swap input values so higher side is series 1
  if a1 < a2
    # swap series parameters
    a1, a2 = a2, a1
    b1, b2 = b2, b1
    slope1, slope2 = slope2, slope1
  end
  puts "get_inputs calling draw arch with params #{a1}, #{b1}, #{slope1}, #{a2}, #{b2}, #{slope2}, #{y_offset}, #{component_name}, #{show_floor_numbers}"
  draw_arch(a1, b1, slope1, a2, b2, slope2, y_offset, component_name, show_floor_numbers)
end

def draw_arch(a1, b1, slope1, a2, b2, slope2, y_offset, component_name, show_floor_numbers)
  
  puts "draw arch called with params"
  
  mod = Sketchup.active_model # Open model
  model_ents = mod.entities # All entities in model
  #sel = mod.selection # Current selection

  points1 = []
  points2 = []



  # Set up components and name them
  ellipse_comp = mod.definitions.add("Arch Section")
  ellipse_comp.name = component_name
  floors_comp = mod.definitions.add("Floor Lines")
  floors_comp.name = ellipse_comp.name + " floors"
  floor_numbers_comp = mod.definitions.add("Floor Numbers") if show_floor_numbers
  floor_numbers_comp.name = ellipse_comp.name  + " floor numbers" if show_floor_numbers
  points1 = []
  # # puts"a1 = #{a1.to_l}, a2 = #{a2.to_l}"

  ### Set initial values ----------------------
  z1 = 0.0
  z2 = 0.0
  a1sq = a1**2
  b1sq = b1**2

  a2sq = a2**2
  b2sq = b2**2
  
  slice_height = 10.feet
  next_slice_height = slice_height

  floor_no = 1
  
  z_level = 0.feet # Height above zero, of current floor
  in_floor_level = 0.feet
  i = 0 # slice counter

  std_floor_height = 10.feet
  non_std_floor_height = {1 => 9.83333.feet, 2 => 15.25.feet, 4 => 15.25.feet, 33 => 15.25.feet, 64 => 15.25.feet, 87  => 20.feet}

  # Set up mesh for arch section
  # Create an empty polygon mesh to add faces to
  
  mesh = Geom::PolygonMesh.new # num_pts, num_poly
  #
  #  #### Iterate over floors  #################################
  #  while z_level <=  a1 do # Keep going until we reach the top - we made sure a1 is the higher apex
  while i <= 120
    # get floor height
    unless non_std_floor_height[floor_no].nil?
      this_floor_height = non_std_floor_height[floor_no]
    else
      this_floor_height = std_floor_height
    end
#    puts " floor_no = #{floor_no}, this_floor_height = #{this_floor_height.to_l}"
    puts "i = #{i}, z_level = #{z_level.to_l}, this_floor_height = #{this_floor_height}, next_slice_height = #{next_slice_height}"
    slice_height = 2.5.feet if i > 85
    
    # Iterate through slices in floor ########################
    while in_floor_level  < this_floor_height do # slice in floor

      # Calculate points to add for current z_level slice(s, and height to stop next slice -----------------------------------
      if this_floor_height < std_floor_height
        this_slice_height = [slice_height, this_floor_height].min
        UI.messagebox "#{i}, this A slice height is zero" if this_slice_height <0.05
      else #this_floor_height >= std_floor_height 
        if this_floor_height - in_floor_level > slice_height
          this_slice_height = slice_height
          UI.messagebox "#{i}, this B slice height is zero" if this_slice_height <0.05
        elsif this_floor_height - in_floor_level > 0.0
          this_slice_height = this_floor_height - in_floor_level 
        else 
          UI.messagebox "Unclear what should happen here"
          this_slice_height = slice_height
        end
      end
      puts "floor_no = #{floor_no} i = #{i}, this_slice_height = #{this_slice_height.to_l}, this_floor_height = #{this_floor_height.to_l}"
      #check whether below, near or above top of lower arch before drawing a slice
      next_z = (z_level + this_slice_height).to_l
      
      if z_level < a2 # we are below the lower apex
        puts "i = #{i}, below lower apex\n"
        case 
        when next_z <= a2 # and so will the next floor be at the estimated this_slice_height
          puts "i = #{i}, next slice below lower apex too\n"
          next_slice_height = [a2 - z_level, this_slice_height].min   
        when next_z > a2 # next slice is above lower apex, and may also be above upper apex if they are closer than this_slice_height apart
          puts "i = #{i}, next slice above lower apex" 
          next_slice_height = [this_slice_height, a1 - z_level, a2 - z_level].min         
        end # case next_z      
      else # z_level > a2 # we are above the lower apex 
        puts "z_level above lower apex"
        case # next_z
        when next_z <= a1 # ... and also below the upper apex with next slice
          puts ".. and also below the upper apex with next slice"
          next_slice_height = this_slice_height        
        when next_z > a1 
          puts "next slice will be above upper apex"
          next_slice_height = [this_slice_height, a1 - z_level].min 
        end
      end # if z_level     

 
      # Add calculated points to mesh
      #      mesh.add_polygon points1[i-1], points2[i-1], points2[i], points1[i] # unless 1 == 0

      # draw floor bottom and set floor labels if required
      # puts " Calling draw_floor for floor  = #{floor_no}"
      #      draw_floor(floor_no, points1[i], points2[i], show_floor_numbers, floors_comp,floor_numbers_comp) if in_floor_level == 0

      puts "Slice #{i}, z_level, this_floor_height, next_slice_height = #{this_floor_height}, #{z_level.to_l}, #{next_slice_height}" 
      # Draw slice
      z1 = z_level
      y1 = z1 * slope1
      z1 = a1 if z1 > a1
      x1 = - b1 * Math.sqrt(1 - z1**2 / a1sq)
      points1[i] = Geom::Point3d.new(x1, y1, z1)

      z2 = z_level
      y2 = z2 * slope1 + y_offset
      z2 >= a2 ? x2 = 0.0 : x2 = - b2 * Math.sqrt(1 - z2**2 / a2sq)
      points2[i]= Geom::Point3d.new(x2, y2, z2)

      ellipse_comp.entities.add_cpoint points1[i]
      ellipse_comp.entities.add_cpoint points2[i]

#      mesh.add_polygon points1[i-1], points2[i-1], points2[i], points1[i] if i > 0
      if next_slice_height <0.05
        UI.messagebox "#{i}, next slice height is zero - breaking out of in-floor loop"
        next_slice_height = slice_height
        break
      end
    
      i += 1 # next slice
      in_floor_level += next_slice_height
      z_level += next_slice_height
  
    end # while in floor
    
if i >= 89
  UI.messagebox "#{i}, floor_no = #{floor_no}, z_level = #{z_level.to_l}, next_slice_height = #{next_slice_height}, in_floor_level = #{in_floor_level.to_l}"

end
      if next_slice_height <0.05
        UI.messagebox "#{i}, next slice height is zero - breaking out of next floor loop" 
         next_slice_height = slice_height
        break
      end
    # Increment floor parameters
    # Next floor
    floor_no += 1
    in_floor_level = 0.feet


#     puts "Next floor_no = #{floor_no.to_s}, z_level = #{z_level.to_l} " if i >=90 and i < 100


  end # do while z_level <= a1 (loop to next floor)
  #

  # Add control points
  ellipse_comp.entities.add_cpoint [0,0,0]
  floors_comp.entities.add_cpoint [0,0,0]
  floor_numbers_comp.entities.add_cpoint [0,0,0] if show_floor_numbers
  if y_offset >0.0
    ellipse_comp.entities.add_cpoint [0, y_offset, 0]
  end
  #   ellipse_comp.entities.add_cpoint [0, a1 * slope1, a1]
  #   ellipse_comp.entities.add_cpoint [0, a2 * slope2, a2]

  # # puts"About to add faces to mesh"
  #  ellipse_comp.entities.add_faces_from_mesh mesh
  # mod.entities.add_faces_from_mesh(mesh, smooth_flag,nil)



  begin
    mod.place_component ellipse_comp  # code that breaks Dynamic Components
  rescue NoMethodError => e
    source = e.backtrace[0]
    pathname = source.split(':').first
    if pathname != __FILE__
      file = File.basename(pathname,".*")
      if file != "dcobservers"
        UI.messagebox "Sorry, error in script " + file.to_s
      end
    else
      # the error is in THIS file !
      UI.messagebox "Sorry, error in DC observers code"
    end
  end
end # draw_arch
